
///    (Brenkman Andrey)
///    9.03.2009 20.1.2010
///    40%


#include "linux-compat.h"

///!
#include "move_list_struct.h"


/// -(1-0)      
#define TEST_H 0 


/// -(1-0)   
///    
///     TEST_WRITE_READ_WN  Go.h
#define TEST_WRITE_READ 0 
                          
                         
/*
Transposition Table
http://chessprogramming.wikispaces.com/Transposition+Table

 A Transposition Table,
 first used in Greenblatt's program Mac Hack [1] [2] , 
 is a database that stores results of previously performed 
 searches. It is a way to greatly reduce the search space 
 of a chess tree with little negative impact. Chess programs, 
 during their brute-force search, encounter the same positions 
 again and again, but from different sequences of moves, which 
 is called a transposition.

 Transposition (and refutation) tables are techniques derived from 
 dynamic programming [3] , a term coined by Richard E. Bellman in 
 the 1950s, when programming meant planning, and dynamic programming 
 was conceived to optimally plan multistage processes [4] .

How it works
 When the search encounters a transposition, it is beneficial to 
 'remember' what was determined last time the position was examined, 
 rather than redoing the entire search again. For this reason, chess 
 programs have a transposition table, which is a large hash table 
 storing information about positions previously searched, how deeply 
 they were searched, and what we concluded about them. Even if the 
 depth (draft) of the related transposition table entry is not big 
 enough, or does not contain the right bound for a cutoff, a best 
 (or good enough) move from a previous search can improve move ordering, 
 and save search time. This is especially true inside an iterative 
 deepening framework, where one gains valuable table hits from 
 previous iterations.

Hash functions
 Hash functions convert chess positions into an almost unique, 
 scalar signature, allowing fast index calculation as well as 
 space saving verification of stored positions.
  Zobrist Hashing


  -
   -    
*/


///       
struct Hash_moves 
{

	///    (  high 32 bits)
	///    64   - (high 32 bits)
	uint32_t key_lock_32_move;

	///      T_list_surplus_moves
	///         32  
	int32_t move; 

	///       
	uint8_t move_delta_depth;

	///   .        
	uint8_t move_age;


	///     ()    ()

	///    (  high 32 bits)
	///    64   - (high 32 bits)
	uint32_t key_lock_32_score;

	/// 
	///       16   
	int32_t score;

	///   (1-  (); 3-  (); 2-  ())
	uint8_t flag_hash_score;

	///       
	uint8_t score_delta_depth;

	///   .        
	uint8_t score_age;

};



namespace Transposition_table
{

///  
//////////////////////////////////////////////////////////////////////////////////////
//private:

#if TEST_WRITE_READ

	///     

	///    
	uint32_t hash_write_all;

	///      
	uint32_t hash_move_write;

	///        (LOWER //   ())
	uint32_t hash_score_write_low;

	///        (UPPER 3 ()   )
	uint32_t hash_score_write_high;


	///   

	///    
	uint32_t hash_read_sorting; 

	///     .  
	uint32_t hash_read_score; 

	///     . LOWER
	uint32_t hash_read_score_LOWER; 

	///     . UPPER 
	uint32_t hash_read_score_UPPER; 

#endif // TEST_WRITE_READ


//////////////////////////////////////////////////////////////////////////////////////
//public:

///   -  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#if TEST_WRITE_READ

	///---------------------------------------------------------------------------
	///  
	void test_hash_ini();

	///---------------------------------------------------------------------------
	///  
	void test_hash_print();

#endif // TEST_WRITE_READ
/// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	///    -  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	///---------------------------------------------------------------------------
	///    -.    
	void hash_size_ini_default
	(
		uint16_t hash_size///  -  
	);

	///---------------------------------------------------------------------------
	///     -       -.
	///    
	void hash_size_resize
	(
		uint16_t hash_size///  -  
	);

	///---------------------------------------------------------------------------
	///     -
	void hash_size_free();

	///---------------------------------------------------------------------------
	///     - 0
	void ini_hash_moves();

	///---------------------------------------------------------------------------
	///         (  )
	void save_best_hash_score_or_move
	(
		uint8_t i,///    
		uint64_t key,/// -
		const struct List & list_surplus_moves,///  
		int32_t score,///  
		uint8_t flag_hash_score,///   (1-  (); 3-  (); 2-  ())
		uint8_t depth,///   
		uint8_t depth_max///   
	);

	///---------------------------------------------------------------------------
	///       .   
	///    
	///        -
	int32_t look_hash_score_white
	(
		const uint64_t key,/// -
		const uint8_t depth,/// 
		const uint8_t depth_max,///  
		const int32_t alpha,///
		const int32_t beta,///
		uint8_t & test_flag_hash_moves,///   (1-  (); 3-  (); 2-  ())
		uint8_t & hash_depth,///  
		uint8_t & hash_depth_max///  
	);

	///---------------------------------------------------------------------------
	///          
	///    
	///        -
	int32_t look_hash_score_black
	(
		const uint64_t key,/// -
		const uint8_t depth,/// 
		const uint8_t depth_max,///  
		const int32_t alpha,///
		const int32_t beta,///
		uint8_t & test_flag_hash_moves,///   (1-  (); 3-  (); 2-  ())
		uint8_t & hash_depth,///  
		uint8_t & hash_depth_max///  
	);

	///---------------------------------------------------------------------------
	///              
	void look_hash_move
	(
		struct List & list_surplus_moves,///  
		const uint64_t key,///  - 
		const uint8_t depth,/// 
		uint8_t  & flag_insert_hash_move///     -     
	);

	///---------------------------------------------------------------------------
	///   
	void set_age_0();

	///---------------------------------------------------------------------------
	///   
	void aging();


//////////////////////////////////////////////////////////////////////////////////////
//private:

	///    -  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	///---------------------------------------------------------------------------
	///       
	inline void save_best_hash_move
	(
		uint8_t i,///    
		uint32_t key_lock_32,///
		const struct List & list_surplus_moves,///  
		uint8_t delta_depth,///   
		uint32_t ind///
	);

	///---------------------------------------------------------------------------
	///   ..    -
	inline void save_hash_move
	(
		uint8_t i,///    
		uint32_t key_lock_32,///
		const struct List & list_surplus_moves,///  
		uint8_t delta_depth,///   
		uint32_t ind///  
	);

	///---------------------------------------------------------------------------
	///      
	inline void save_best_hash_score
	(
		uint32_t key_lock_32,///
		int32_t score,///  
		uint8_t flag_hash_score,///   (1-  (); 3-  (); 2-  ())
		uint8_t delta_depth,///   
		uint32_t ind///
	);

	///---------------------------------------------------------------------------
	///   ..    -
	inline void save_hash_score
	(
		uint32_t key_lock_32,///
		int32_t score,///  
		uint8_t flag_hash_score,///   (1-  (); 3-  (); 2-  ())
		uint8_t delta_depth,///   
		uint32_t ind///  
	);

	///---------------------------------------------------------------------------
	///  -.     
	///     
	void TEST_Hash_index_print();


//////////////////////////////////////////////////////////////////////////////////////
//public:

	///---------------------------------------------------------------------------
	///  -          
	void insert_hash_moves
	(
		struct List & list,///   
		const int32_t hash_moves,//
		uint8_t  & flag_insert_hash_move///     -     
	);


};
